{-# LANGUAGE DataKinds #-}
{-# LANGUAGE DeriveGeneric #-}
{-# LANGUAGE TypeOperators #-}
{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE TemplateHaskell #-}

module HAL.Client where

import Data.Proxy
import GHC.Generics
import Servant.API
import Servant.Client hiding (Response)
import Data.Text
import Data.Map
import Data.Aeson
import qualified Data.ByteString.Lazy as BSL
import qualified Codec.Binary.UTF8.String as UTF

import Control.Lens as L (makeLenses)

type HALAPI = Search :<|> Structure

type Search = "search"
  -- fl determine which fields will be returned it can be a list of fields or *
  :> QueryParams "fl" Text
  -- TODO: type this monster
  -- fq is to filter request
  :> QueryParams "fq" Text
  -- pretty much clear, (Asc || Desc) + field you want to sort by
  :> QueryParam "sort" SortField
  -- permit to start a the x result
  :> QueryParam "start" Int
  -- use rows to make the request only return the x number of result
  :> QueryParam "rows" Int
  :> Get '[JSON] Response

type Structure = "ref" :> "structure"
  :> QueryParam "fq" Text
  :> QueryParam "fl" Text
  :> QueryParam "rows" Int
  :> Get '[JSON] Response


-- Get's argument type
data SortField = Asc Text | Desc Text
  deriving (Show)
instance ToHttpApiData SortField where
  toUrlPiece (Asc t) = t <> " asc"
  toUrlPiece (Desc t) = t <> " desc"

asc :: Text -> Maybe SortField
asc = Just . Asc

desc :: Text -> Maybe SortField
desc = Just . Desc

newtype Doc = Doc (Map Text Value)
  deriving (Generic)

instance Eq Doc where
  (==) (Doc doc) (Doc doc') = (doc ! "docid") == (doc' ! "docid")

instance Show Doc where
  show (Doc o) = (UTF.decode $ BSL.unpack $ encode $ o ! "label_s")
    <> "("
    <> (show . encode $ o ! "docid")
    <> ")"

instance FromJSON Doc
instance ToJSON Doc

-- Response type
data Response = Response
  {
   _numFound :: Integer,
   _start :: Int,
   _docs :: [Doc]
  } deriving (Show, Generic)
L.makeLenses ''Response


instance FromJSON Response where
  parseJSON (Object o) = Response <$>
                         ((o .: "response") >>= (.: "numFound"))
                         <*> ((o .: "response") >>= (.: "start"))
                         <*> ((o .: "response") >>= (.: "docs"))


halAPI :: Proxy HALAPI
halAPI = Proxy

structure :: Maybe Text -- fq
  -> Maybe Text
  -> Maybe Int -- rows
  -> ClientM Response

-- search should always have at least `docid` and `label_s` in his fl params
search :: [Text] -- fl
  -> [Text] -- fq
  -> Maybe SortField -- sort
  -> Maybe Int -- start
  -> Maybe Int -- rows
  -> ClientM Response
(search :<|> structure) = client halAPI
